@@ -14,6 +14,7 @@
 
 #include "../../config.h"
 #include "inotifytools/inotifytools.h"
+#include "inotifytools_p.h"
 
 #include <string.h>
 #include <strings.h>
@@ -27,6 +28,8 @@
 #include <unistd.h>
 #include <dirent.h>
 #include <time.h>
+#include <regex.h>
+#include <setjmp.h>
 
 #include "inotifytools/inotify.h"
 
@@ -115,14 +118,9 @@
  *       different filenames, and when an event occurs on the inode, generate an
  *       event for each filename.
  *
- * @todo Improve wd/filename storage.  At the moment, wds start begin allocated
- *       from 1 and grow upwards, and don't go down after watches are released.
- *       The current strategy in libinotifytools is to simply use the wd as an
- *       array index.  This is obviously buggy; if you have a thousand watches,
- *       remove them all, and add one more watch, we'll end up with an array
- *       with about a thousand empty spots followed by one used spot!  We need
- *       a proper solution which doesn't rely at all on the kernel policy of
- *       allocating wds.  Some kind of hopefully efficient lookup table.
+ * @todo Improve wd/filename storage.  At the moment a doubly-linked list is used,
+ *       so lookups take linear time.  This could be a problem if watching
+ *       hundreds of thousands of files.
  */
 
 #define MAX_EVENTS 4096
@@ -133,44 +131,29 @@
 #define INSTANCES_PATH    INOTIFY_PROCDIR "max_user_instances"
 
 static int inotify_fd;
-static unsigned int * hit_access = 0;
-static unsigned int * hit_modify = 0;
-static unsigned int * hit_attrib = 0;
-static unsigned int * hit_close_nowrite = 0;
-static unsigned int * hit_close_write = 0;
-static unsigned int * hit_open = 0;
-static unsigned int * hit_move_self = 0;
-static unsigned int * hit_moved_from = 0;
-static unsigned int * hit_moved_to = 0;
-static unsigned int * hit_create = 0;
-static unsigned int * hit_delete = 0;
-static unsigned int * hit_delete_self = 0;
-static unsigned int * hit_unmount = 0;
-static unsigned int * hit_total = 0;
-static unsigned int num_access;
-static unsigned int num_modify;
-static unsigned int num_attrib;
-static unsigned int num_close_nowrite;
-static unsigned int num_close_write;
-static unsigned int num_open;
-static unsigned int num_move_self;
-static unsigned int num_moved_to;
-static unsigned int num_moved_from;
-static unsigned int num_create;
-static unsigned int num_delete;
-static unsigned int num_delete_self;
-static unsigned int num_unmount;
-static unsigned int num_total;
+static unsigned  num_access;
+static unsigned  num_modify;
+static unsigned  num_attrib;
+static unsigned  num_close_nowrite;
+static unsigned  num_close_write;
+static unsigned  num_open;
+static unsigned  num_move_self;
+static unsigned  num_moved_to;
+static unsigned  num_moved_from;
+static unsigned  num_create;
+static unsigned  num_delete;
+static unsigned  num_delete_self;
+static unsigned  num_unmount;
+static unsigned  num_total;
 static int collect_stats = 0;
-static int WATCHES_SIZE;
-static char ** watches;
-static int error;
-static unsigned int num_watches;
-static unsigned int watches_buf_size;
-static int init = 0;
-static char * timefmt = 0;
 
-void allocatestatmem( int from, int to );
+struct rbtree *tree_wd = 0;
+struct rbtree *tree_filename = 0;
+static int error = 0;
+static int init = 0;
+static char* timefmt = 0;
+static regex_t* regex = 0;
+
 int isdir( char const * path );
 void record_stats( struct inotify_event const * event );
 int onestr_to_event(char const * event);
@@ -192,9 +175,11 @@ int onestr_to_event(char const * event);
  *             "input not in required range!");
  * @endcode
  */
-#define niceassert(cond,mesg) _niceassert((int)cond, __LINE__, __FILE__, \
+#define niceassert(cond,mesg) _niceassert((long)cond, __LINE__, __FILE__, \
                                           #cond, mesg)
 
+#define nasprintf(...) niceassert( -1 != asprintf(__VA_ARGS__), "out of memory")
+
 /**
  * @internal
  * Assert that a condition evaluates to true, and optionally output a message
@@ -212,7 +197,7 @@ int onestr_to_event(char const * event);
  *
  * @param  mesg  A human-readable error message shown if assertion fails.
  */
-void _niceassert( int cond, int line, char const * file, char const * condstr,
+void _niceassert( long cond, int line, char const * file, char const * condstr,
                   char const * mesg ) {
 	if ( cond ) return;
 
@@ -225,46 +210,6 @@ void _niceassert( int cond, int line, char const * file, char const * condstr,
 	}
 }
 
-/**
- * @internal
- * Resize a block of memory and, if enlarged, fill the new area with zeroes.
- *
- * @param ptr pointer to memory to resize; can be NULL or a pointer returned by
- *            @a malloc, @a realloc or @a calloc
- *
- * @param from Original size of memory.
- *
- * @param to   New size of memory.
- *
- * @return A pointer to a block of memory of size @a to, or NULL.  If a block
- *         of memory was enlarged, the newly allocated space will be set to 0.
- */
-void * recalloc( void * ptr, size_t from, size_t to ) {
-	static void * ret;
-	ret = realloc( ptr, to );
-	niceassert( ret, "out of memory" );
-	if ( ret && to > from ) {
-		memset( ret + from, 0, (to - from) );
-	}
-	return ret;
-}
-
-/**
- * @internal
- * Resize the watches buffer, and the stats buffers if stats are enabled.
- *
- * @param diff Amount of space (in elements) to add or subtract to buffers.
- */
-void resize_buffers( int diff ) {
-	static int oldsize;
-	oldsize = watches_buf_size;
-	watches_buf_size += diff;
-	watches = (char **)recalloc( watches, oldsize*sizeof(char *),
-	                             watches_buf_size*sizeof(char *) );
-	if ( collect_stats ) allocatestatmem( oldsize*sizeof(char *),
-	                                      watches_buf_size*sizeof(char *) );
-}
-
 /**
  * @internal
  * Construct a string from a character.
@@ -300,6 +245,34 @@ int read_num_from_file( char * filename, int * num ) {
 	return 1;
 }
 
+int wd_compare(const void *d1, const void *d2, const void *config) {
+	if (!d1 || !d2) return d1 - d2;
+	return ((watch*)d1)->wd - ((watch*)d2)->wd;
+}
+
+int filename_compare(const void *d1, const void *d2, const void *config) {
+	if (!d1 || !d2) return d1 - d2;
+	return strcmp(((watch*)d1)->filename, ((watch*)d2)->filename);
+}
+
+/**
+ * @internal
+ */
+watch *watch_from_wd( int wd ) {
+	watch w;
+	w.wd = wd;
+	return (watch*)rbfind(&w, tree_wd);
+}
+
+/**
+ * @internal
+ */
+watch *watch_from_filename( char const *filename ) {
+	watch w;
+	w.filename = (char*)filename;
+	return (watch*)rbfind(&w, tree_filename);
+}
+
 /**
  * Initialise inotify.
  *
@@ -310,6 +283,8 @@ int read_num_from_file( char * filename, int * num ) {
  *         obtained from inotifytools_error().
  */
 int inotifytools_initialize() {
+	if (init) return 1;
+
 	error = 0;
 	// Try to initialise inotify
 	inotify_fd = inotify_init();
@@ -318,44 +293,120 @@ int inotifytools_initialize() {
 		return 0;
 	}
 
-	if ( !read_num_from_file( WATCHES_SIZE_PATH, &WATCHES_SIZE ) ) return 0;
-
-	watches_buf_size = WATCHES_SIZE;
-	watches = (char **)calloc( watches_buf_size, sizeof(char *) );
-	niceassert( watches, "out of memory, gosh darn it" );
-
 	collect_stats = 0;
-	num_watches = 0;
 	init = 1;
+	tree_wd = rbinit(wd_compare, 0);
+	tree_filename = rbinit(filename_compare, 0);
+	timefmt = 0;
 
 	return 1;
 }
 
 /**
  * @internal
- * Allocate memory for storing statistics.
- *
- * @param from previous size of stat buffers
- *
- * @param to new size of stat buffers
  */
-void allocatestatmem( int from, int to ) {
-	hit_access = (unsigned int *)recalloc( hit_access, from, to );
-	hit_modify = (unsigned int *)recalloc( hit_modify, from, to );
-	hit_attrib = (unsigned int *)recalloc( hit_attrib, from, to );
-	hit_close_nowrite = (unsigned int *)recalloc( hit_close_nowrite, from, to );
-	hit_close_write = (unsigned int *)recalloc( hit_close_write, from, to );
-	hit_open = (unsigned int *)recalloc( hit_open, from, to );
-	hit_move_self = (unsigned int *)recalloc( hit_move_self, from, to );
-	hit_moved_from = (unsigned int *)recalloc( hit_moved_from, from, to );
-	hit_moved_to = (unsigned int *)recalloc( hit_moved_to, from, to );
-	hit_create = (unsigned int *)recalloc( hit_create, from, to );
-	hit_delete = (unsigned int *)recalloc( hit_delete, from, to );
-	hit_delete_self = (unsigned int *)recalloc( hit_delete_self, from, to );
-	hit_unmount = (unsigned int *)recalloc( hit_unmount, from, to );
-	hit_total = (unsigned int *)recalloc( hit_total, from, to );
+void destroy_watch(watch *w) {
+	if (w->filename) free(w->filename);
+	free(w);
 }
 
+/**
+ * @internal
+ */
+void cleanup_tree(const void *nodep,
+                 const VISIT which,
+                 const int depth, void* arg) {
+	if (which != endorder && which != leaf) return;
+	watch *w = (watch*)nodep;
+	destroy_watch(w);
+}
+
+/**
+ * Close inotify and free the memory used by inotifytools.
+ *
+ * If you call this function, you must call inotifytools_initialize()
+ * again before any other functions can be used.
+ */
+void inotifytools_cleanup() {
+	if (!init) return;
+
+	init = 0;
+	close(inotify_fd);
+	collect_stats = 0;
+	error = 0;
+	timefmt = 0;
+
+	if (regex) {
+		regfree(regex);
+		free(regex);
+		regex = 0;
+	}
+
+	rbwalk(tree_wd, cleanup_tree, 0);
+	rbdestroy(tree_wd); tree_wd = 0;
+	rbdestroy(tree_filename); tree_filename = 0;
+}
+
+/**
+ * @internal
+ */
+void empty_stats(const void *nodep,
+                 const VISIT which,
+                 const int depth, void *arg) {
+    if (which != endorder && which != leaf) return;
+	watch *w = (watch*)nodep;
+	w->hit_access = 0;
+	w->hit_modify = 0;
+	w->hit_attrib = 0;
+	w->hit_close_nowrite = 0;
+	w->hit_close_write = 0;
+	w->hit_open = 0;
+	w->hit_move_self = 0;
+	w->hit_moved_from = 0;
+	w->hit_moved_to = 0;
+	w->hit_create = 0;
+	w->hit_delete = 0;
+	w->hit_delete_self = 0;
+	w->hit_unmount = 0;
+	w->hit_total = 0;
+}
+
+/**
+ * @internal
+ */
+void replace_filename(const void *nodep,
+                      const VISIT which,
+                      const int depth, void *arg) {
+    if (which != endorder && which != leaf) return;
+	watch *w = (watch*)nodep;
+	char *old_name = ((char**)arg)[0];
+	char *new_name = ((char**)arg)[1];
+	int old_len = *((int*)&((char**)arg)[2]);
+	char *name;
+	if ( 0 == strncmp( old_name, w->filename, old_len ) ) {
+		nasprintf( &name, "%s%s", new_name, &(w->filename[old_len]) );
+		if (!strcmp( w->filename, new_name )) {
+			free(name);
+		} else {
+			rbdelete(w, tree_filename);
+			free( w->filename );
+			w->filename = name;
+			rbsearch(w, tree_filename);
+		}
+	}
+}
+
+/**
+ * @internal
+ */
+void get_num(const void *nodep,
+             const VISIT which,
+             const int depth, void *arg) {
+    if (which != endorder && which != leaf) return;
+	++(*((int*)arg));
+}
+
+
 /**
  * Initialize or reset statistics.
  *
@@ -371,26 +422,9 @@ void allocatestatmem( int from, int to ) {
 void inotifytools_initialize_stats() {
 	niceassert( init, "inotifytools_initialize not called yet" );
 
-	// if not already collecting stats, allocate buffer
-	if ( !collect_stats ) {
-		allocatestatmem( 0, watches_buf_size );
-	}
 	// if already collecting stats, reset stats
-	else {
-		memset( hit_access, 0, watches_buf_size );
-		memset( hit_modify, 0, watches_buf_size );
-		memset( hit_attrib, 0, watches_buf_size );
-		memset( hit_close_nowrite, 0, watches_buf_size );
-		memset( hit_close_write, 0, watches_buf_size );
-		memset( hit_open, 0, watches_buf_size );
-		memset( hit_move_self, 0, watches_buf_size );
-		memset( hit_moved_from, 0, watches_buf_size );
-		memset( hit_moved_to, 0, watches_buf_size );
-		memset( hit_create, 0, watches_buf_size );
-		memset( hit_delete, 0, watches_buf_size );
-		memset( hit_delete_self, 0, watches_buf_size );
-		memset( hit_unmount, 0, watches_buf_size );
-		memset( hit_total, 0, watches_buf_size );
+	if (collect_stats) {
+		rbwalk(tree_wd, empty_stats, 0);
 	}
 
 	num_access = 0;
@@ -711,7 +745,7 @@ char * inotifytools_event_to_str_sep(int events, char sep)
 }
 
 /**
- * Get the filename for a given watch descriptor.
+ * Get the filename used to establish a watch.
  *
  * inotifytools_initialize() must be called before this function can
  * be used.
@@ -727,11 +761,14 @@ char * inotifytools_event_to_str_sep(int events, char sep)
  *       Also, this is not necessarily the filename which might have been used
  *       to cause an event on the file, since inotify is inode based and there
  *       can be many filenames mapping to a single inode.
+ *       Finally, if a file is moved or renamed while being watched, the
+ *       filename returned will still be the original name.
  */
 char * inotifytools_filename_from_wd( int wd ) {
-	if ( wd <= 0 ) return 0;
 	niceassert( init, "inotifytools_initialize not called yet" );
-	return watches[wd-1];
+	watch *w = watch_from_wd(wd);
+	if (!w) return 0;
+	return w->filename;
 }
 
 /**
@@ -744,30 +781,119 @@ char * inotifytools_filename_from_wd( int wd ) {
  *
  * @return watch descriptor associated with filename, or -1 if @a filename is
  *         not associated with any watch descriptor.
+ *
+ * @note The filename specified must always be the original name used to
+ *       establish the watch.
  */
 int inotifytools_wd_from_filename( char const * filename ) {
-	niceassert( filename, 0 );
 	niceassert( init, "inotifytools_initialize not called yet" );
-	static int i;
-	for ( i = 0; i < WATCHES_SIZE; ++i ) {
-		if ( watches[i] && 0 == strcmp( filename, watches[i] ) ) {
-			return i+1;
-		}
-	}
-	return -1;
+	watch *w = watch_from_filename(filename);
+	if (!w) return -1;
+	return w->wd;
 }
 
 /**
- * Remove a watch on a file specified by filename.
+ * Set the filename for a particular watch descriptor.
  *
- * @param filename Name of file on which watch should be removed.
+ * This function should be used to update a filename when a file is known to
+ * have been moved or renamed.  At the moment, libinotifytools does not
+ * automatically handle this situation.
  *
- * @return 1 on success, 0 on failure.  On failure, the error can be
- *         obtained from inotifytools_error().
+ * inotifytools_initialize() must be called before this function can
+ * be used.
+ *
+ * @param wd Watch descriptor.
+ *
+ * @param filename New filename.
  */
-int inotifytools_remove_watch_by_filename( char const * filename ) {
-	return inotifytools_remove_watch_by_wd( inotifytools_wd_from_filename(
-	       filename ) );
+void inotifytools_set_filename_by_wd( int wd, char const * filename ) {
+	niceassert( init, "inotifytools_initialize not called yet" );
+	watch *w = watch_from_wd(wd);
+	if (!w) return;
+	if (w->filename) free(w->filename);
+	w->filename = strdup(filename);
+}
+
+/**
+ * Set the filename for one or more watches with a particular existing filename.
+ *
+ * This function should be used to update a filename when a file is known to
+ * have been moved or renamed.  At the moment, libinotifytools does not
+ * automatically handle this situation.
+ *
+ * inotifytools_initialize() must be called before this function can
+ * be used.
+ *
+ * @param oldname Current filename.
+ *
+ * @param newname New filename.
+ */
+void inotifytools_set_filename_by_filename( char const * oldname,
+                                            char const * newname ) {
+	watch *w = watch_from_filename(oldname);
+	if (!w) return;
+	if (w->filename) free(w->filename);
+	w->filename = strdup(newname);
+}
+
+/**
+ * Replace a certain filename prefix on all watches.
+ *
+ * This function should be used to update filenames for an entire directory tree
+ * when a directory is known to have been moved or renamed.  At the moment,
+ * libinotifytools does not automatically handle this situation.
+ *
+ * inotifytools_initialize() must be called before this function can
+ * be used.
+ *
+ * @param oldname Current filename prefix.
+ *
+ * @param newname New filename prefix.
+ *
+ * @section example Example
+ * @code
+ * // if /home/user1/original_dir is moved to /home/user2/new_dir, then to
+ * // update all watches:
+ * inotifytools_replace_filename( "/home/user1/original_dir",
+ *                                "/home/user2/new_dir" );
+ * @endcode
+ */
+void inotifytools_replace_filename( char const * oldname,
+                                    char const * newname ) {
+	if ( !oldname || !newname ) return;
+	char *names[2+sizeof(int)/sizeof(char*)];
+	names[0] = (char*)oldname;
+	names[1] = (char*)newname;
+	*((int*)&names[2]) = strlen(oldname);
+	rbwalk(tree_filename, replace_filename, (void*)names);
+}
+
+/**
+ * @internal
+ */
+int remove_inotify_watch(watch *w) {
+	error = 0;
+	int status = inotify_rm_watch( inotify_fd, w->wd );
+	if ( status < 0 ) {
+		fprintf(stderr, "Failed to remove watch on %s: %s\n", w->filename,
+		        strerror(status) );
+		error = status;
+		return 0;
+	}
+	return 1;
+}
+
+/**
+ * @internal
+ */
+watch *create_watch(int wd, char *filename) {
+	if ( wd <= 0 || !filename) return 0;
+
+	watch *w = (watch*)calloc(1, sizeof(watch));
+	w->wd = wd;
+	w->filename = strdup(filename);
+	rbsearch(w, tree_wd);
+	rbsearch(w, tree_filename);
 }
 
 /**
@@ -778,28 +904,42 @@ int inotifytools_remove_watch_by_filename( char const * filename ) {
  *
  * @param wd Watch descriptor of watch to be removed.
  *
- * @return 1 on success, 0 on failure.  On failure, the error can be
+ * @return 1 on success, 0 on failure.  If the given watch doesn't exist,
+ *         returns 1.  On failure, the error can be
  *         obtained from inotifytools_error().
  */
 int inotifytools_remove_watch_by_wd( int wd ) {
-	niceassert( wd > 0, 0 );
-	niceassert( watches[wd - 1], "No watch exists for requested wd" );
 	niceassert( init, "inotifytools_initialize not called yet" );
-	error = 0;
-	static int status;
-	status = inotify_rm_watch( inotify_fd, wd );
-	if ( status < 0 ) {
-		fprintf(stderr, "Failed to remove watch on %s: %s\n", watches[wd - 1],
-		        strerror(status) );
-		error = status;
-		return 0;
-	}
-	free( watches[wd - 1] );
-	watches[wd - 1] = NULL;
-	num_watches--;
-	if ( (int)num_watches < (int)watches_buf_size - (int)WATCHES_SIZE - 1 ) {
-		resize_buffers( -WATCHES_SIZE );
-	}
+	watch *w = watch_from_wd(wd);
+	if (!w) return 1;
+
+	if (!remove_inotify_watch(w)) return 0;
+	rbdelete(w, tree_wd);
+	rbdelete(w, tree_filename);
+	destroy_watch(w);
+	return 1;
+}
+
+/**
+ * Remove a watch on a file specified by filename.
+ *
+ * @param filename Name of file on which watch should be removed.
+ *
+ * @return 1 on success, 0 on failure.  On failure, the error can be
+ *         obtained from inotifytools_error().
+ *
+ * @note The filename specified must always be the original name used to
+ *       establish the watch.
+ */
+int inotifytools_remove_watch_by_filename( char const * filename ) {
+	niceassert( init, "inotifytools_initialize not called yet" );
+	watch *w = watch_from_filename(filename);
+	if (!w) return 1;
+
+	if (!remove_inotify_watch(w)) return 0;
+	rbdelete(w, tree_wd);
+	rbdelete(w, tree_filename);
+	destroy_watch(w);
 	return 1;
 }
 
@@ -857,18 +997,17 @@ int inotifytools_watch_files( char const * filenames[], int events ) {
 			} // else
 		} // if ( wd < 0 )
 
+		char *filename;
 		// Always end filename with / if it is a directory
 		if ( !isdir(filenames[i])
 		     || filenames[i][strlen(filenames[i])-1] == '/') {
-			watches[wd - 1] = strdup(filenames[i]);
+			filename = strdup(filenames[i]);
 		}
 		else {
-			niceassert(-1!=asprintf( &watches[wd - 1], "%s/", filenames[i] ),0);
-		}
-		num_watches++;
-		if ( num_watches == watches_buf_size ) {
-			resize_buffers( WATCHES_SIZE );
+			nasprintf( &filename, "%s/", filenames[i] );
 		}
+		create_watch(wd, filename);
+		free(filename);
 	} // for
 
 	return 1;
@@ -894,6 +1033,11 @@ int inotifytools_watch_files( char const * filenames[], int events ) {
  *       inotifytools_next_events() frequently; between calls to this function,
  *       inotify events will be queued in the kernel, and eventually the queue
  *       will overflow and you will miss some events.
+ *
+ * @note If the function inotifytools_ignore_events_by_regex() has been called
+ *       with a non-NULL parameter, this function will not return on events
+ *       which match the regular expression passed to that function.  However,
+ *       the @a timeout period begins again each time a matching event occurs.
  */
 struct inotify_event * inotifytools_next_event( int timeout ) {
 	return inotifytools_next_events( timeout, 1 );
@@ -943,6 +1087,11 @@ struct inotify_event * inotifytools_next_event( int timeout ) {
  *       inotifytools_next_events() frequently; between calls to this function,
  *       inotify events will be queued in the kernel, and eventually the queue
  *       will overflow and you will miss some events.
+ *
+ * @note If the function inotifytools_ignore_events_by_regex() has been called
+ *       with a non-NULL parameter, this function will not return on events
+ *       which match the regular expression passed to that function.  However,
+ *       the @a timeout period begins again each time a matching event occurs.
  */
 struct inotify_event * inotifytools_next_events( int timeout, int num_events ) {
 	niceassert( init, "inotifytools_initialize not called yet" );
@@ -954,12 +1103,29 @@ struct inotify_event * inotifytools_next_events( int timeout, int num_events ) {
 	static struct inotify_event * ret;
 	static int first_byte = 0;
 	static ssize_t bytes;
+	static jmp_buf jmp;
+	static char match_name[MAX_STRLEN];
+
+#define RETURN(A) {\
+	if (regex) {\
+		inotifytools_snprintf(match_name, MAX_STRLEN, A, "%w%f");\
+		if (0 == regexec(regex, match_name, 0, 0, 0)) {\
+			longjmp(jmp,0);\
+		}\
+	}\
+	if ( collect_stats ) {\
+		record_stats( A );\
+	}\
+	return A;\
+}
+
+	setjmp(jmp);
 
 	error = 0;
 
 	// first_byte is index into event buffer
 	if ( first_byte != 0
-	  && first_byte <= bytes - sizeof(struct inotify_event) ) {
+	  && first_byte <= (int)(bytes - sizeof(struct inotify_event)) ) {
 
 		ret = (struct inotify_event *)((char *)&event[0] + first_byte);
 		first_byte += sizeof(struct inotify_event) + ret->len;
@@ -975,19 +1141,17 @@ struct inotify_event * inotifytools_next_events( int timeout, int num_events ) {
 			// read remainder.
 			// oh, and they BETTER NOT overlap.
 			// Boy I hope this code works.
-			niceassert( (unsigned)((char *)&event[0] +
+			// But I think this can never happen due to how inotify is written.
+			niceassert( (long)((char *)&event[0] +
 			            sizeof(struct inotify_event) +
-			            event[0].len) <= (unsigned)ret,
+			            event[0].len) <= (long)ret,
 			            "extremely unlucky user, death imminent" );
 			// how much of the event do we have?
 			bytes = (char *)&event[0] + bytes - (char *)ret;
 			memcpy( &event[0], ret, bytes );
 			return inotifytools_next_events( timeout, num_events );
 		}
-		if ( collect_stats ) {
-			record_stats( ret );
-		}
-		return ret;
+		RETURN(ret);
 
 	}
 
@@ -997,7 +1161,7 @@ struct inotify_event * inotifytools_next_events( int timeout, int num_events ) {
 
 
 	static ssize_t this_bytes;
-	static int bytes_to_read;
+	static unsigned int bytes_to_read;
 	static int rc;
 	static fd_set read_fds;
 
@@ -1043,9 +1207,6 @@ struct inotify_event * inotifytools_next_events( int timeout, int num_events ) {
 		                "events occurred at once.\n");
 		return NULL;
 	}
-	if ( this_bytes < bytes_to_read ) {
-		fprintf(stderr, "Warning, too few bytes read from inotify!\n" );
-	}
 	bytes += this_bytes;
 
 	ret = &event[0];
@@ -1055,10 +1216,10 @@ struct inotify_event * inotifytools_next_events( int timeout, int num_events ) {
 	if ( first_byte == bytes ) {
 		first_byte = 0;
 	}
-	if ( collect_stats ) {
-		record_stats( ret );
-	}
-	return ret;
+
+	RETURN(ret);
+
+#undef RETURN
 }
 
 /**
@@ -1087,6 +1248,43 @@ struct inotify_event * inotifytools_next_events( int timeout, int num_events ) {
  *       as to whether or not those files will be watched.
  */
 int inotifytools_watch_recursively( char const * path, int events ) {
+	return inotifytools_watch_recursively_with_exclude( path, events, 0 );
+}
+
+/**
+ * Set up recursive watches on an entire directory tree, optionally excluding
+ * some directories.
+ *
+ * inotifytools_initialize() must be called before this function can
+ * be used.
+ *
+ * @author UH
+ *
+ * @param path path of directory or file to watch.  If the path is a directory,
+ *             every subdirectory will also be watched for the same events up
+ *             to the maximum readable depth.  If the path is a file, the file
+ *             is watched exactly as if inotifytools_watch_file() were used.
+ *
+ * @param exclude_list NULL terminated path list of directories not to watch.
+ *                     Can be NULL if no paths are to be excluded.
+ *                     Directories may or may not include a trailing '/'.
+ *
+ * @param events Inotify events to watch for.  See section \ref events.
+ *
+ * @return 1 on success, 0 on failure.  On failure, the error can be
+ *         obtained from inotifytools_error().  Note that some errors on
+ *         subdirectories will be ignored; for example, if you watch a directory
+ *         tree which contains some directories which you do not have access to,
+ *         those directories will not be watched, but this function will still
+ *         return 1 if no other errors occur.
+ *
+ * @note This function does not attempt to work atomically.  If you use this
+ *       function to watch a directory tree and files or directories are being
+ *       created or removed within that directory tree, there are no guarantees
+ *       as to whether or not those files will be watched.
+ */
+int inotifytools_watch_recursively_with_exclude( char const * path, int events,
+                                                 char const ** exclude_list ) {
 	niceassert( init, "inotifytools_initialize not called yet" );
 
 	DIR * dir;
@@ -1105,7 +1303,7 @@ int inotifytools_watch_recursively( char const * path, int events ) {
 	}
 
 	if ( path[strlen(path)-1] != '/' ) {
-		niceassert( -1 != asprintf( &my_path, "%s/", path ), 0 );
+		nasprintf( &my_path, "%s/", path );
 	}
 	else {
 		my_path = (char *)path;
@@ -1119,8 +1317,7 @@ int inotifytools_watch_recursively( char const * path, int events ) {
 	while ( ent ) {
 		if ( (0 != strcmp( ent->d_name, "." )) &&
 		     (0 != strcmp( ent->d_name, ".." )) ) {
-			niceassert( -1 != asprintf(&next_file,"%s%s", my_path, ent->d_name),
-			            0 );
+			nasprintf(&next_file,"%s%s", my_path, ent->d_name);
 			if ( -1 == lstat64( next_file, &my_stat ) ) {
 				error = errno;
 				free( next_file );
@@ -1134,20 +1331,43 @@ int inotifytools_watch_recursively( char const * path, int events ) {
 			else if ( S_ISDIR( my_stat.st_mode ) &&
 			          !S_ISLNK( my_stat.st_mode )) {
 				free( next_file );
-				niceassert( -1 != asprintf(&next_file,"%s%s/", my_path,
-				                           ent->d_name), 0 );
-				static int status;
-				status = inotifytools_watch_recursively( next_file, events );
-				// For some errors, we will continue.
-				if ( !status && (EACCES != error) && (ENOENT != error) &&
-				     (ELOOP != error) ) {
-					free( next_file );
-					if ( my_path != path ) free( my_path );
-					closedir( dir );
-					return 0;
+				nasprintf(&next_file,"%s%s/", my_path, ent->d_name);
+				static unsigned int no_watch;
+				static char const ** exclude_entry;
+
+				no_watch = 0;
+				for (exclude_entry = exclude_list;
+					 exclude_entry && *exclude_entry && !no_watch;
+					 ++exclude_entry) {
+					static int exclude_length;
+
+					exclude_length = strlen(*exclude_entry);
+					if ((*exclude_entry)[exclude_length-1] == '/') {
+						--exclude_length;
+					}
+					if ( strlen(next_file) == (unsigned)(exclude_length + 1) &&
+					    !strncmp(*exclude_entry, next_file, exclude_length)) {
+						// directory found in exclude list
+						no_watch = 1;
+					}
 				}
+				if (!no_watch) {
+					static int status;
+					status = inotifytools_watch_recursively_with_exclude(
+					              next_file,
+					              events,
+					              exclude_list );
+					// For some errors, we will continue.
+					if ( !status && (EACCES != error) && (ENOENT != error) &&
+					     (ELOOP != error) ) {
+						free( next_file );
+						if ( my_path != path ) free( my_path );
+						closedir( dir );
+						return 0;
+					}
+				} // if !no_watch
 				free( next_file );
-			}
+			} // if isdir and not islnk
 			else {
 				free( next_file );
 			}
@@ -1156,75 +1376,111 @@ int inotifytools_watch_recursively( char const * path, int events ) {
 		error = 0;
 	}
 
-	if ( my_path != path ) free( my_path );
 	closedir( dir );
 
-	return inotifytools_watch_file( path, events );
+	int ret = inotifytools_watch_file( my_path, events );
+	if ( my_path != path ) free( my_path );
+        return ret;
 }
 
 /**
  * @internal
  */
 void record_stats( struct inotify_event const * event ) {
-
+	if (!event) return;
+	watch *w = watch_from_wd(event->wd);
+	if (!w) return;
 	if ( IN_ACCESS & event->mask ) {
-		hit_access[event->wd - 1]++;
-		num_access++;
+		++w->hit_access;
+		++num_access;
 	}
 	if ( IN_MODIFY & event->mask ) {
-		hit_modify[event->wd - 1]++;
-		num_modify++;
+		++w->hit_modify;
+		++num_modify;
 	}
 	if ( IN_ATTRIB & event->mask ) {
-		hit_attrib[event->wd - 1]++;
-		num_attrib++;
+		++w->hit_attrib;
+		++num_attrib;
 	}
 	if ( IN_CLOSE_WRITE & event->mask ) {
-		hit_close_write[event->wd - 1]++;
-		num_close_write++;
+		++w->hit_close_write;
+		++num_close_write;
 	}
 	if ( IN_CLOSE_NOWRITE & event->mask ) {
-		hit_close_nowrite[event->wd - 1]++;
-		num_close_nowrite++;
+		++w->hit_close_nowrite;
+		++num_close_nowrite;
 	}
 	if ( IN_OPEN & event->mask ) {
-		hit_open[event->wd - 1]++;
-		num_open++;
+		++w->hit_open;
+		++num_open;
 	}
 	if ( IN_MOVED_FROM & event->mask ) {
-		hit_moved_from[event->wd - 1]++;;
-		num_moved_from++;
+		++w->hit_moved_from;
+		++num_moved_from;
 	}
 	if ( IN_MOVED_TO & event->mask ) {
-		hit_moved_to[event->wd - 1]++;
-		num_moved_to++;
+		++w->hit_moved_to;
+		++num_moved_to;
 	}
 	if ( IN_CREATE & event->mask ) {
-		hit_create[event->wd - 1]++;
-		num_create++;
+		++w->hit_create;
+		++num_create;
 	}
 	if ( IN_DELETE & event->mask ) {
-		hit_delete[event->wd - 1]++;
-		num_delete++;
+		++w->hit_delete;
+		++num_delete;
 	}
 	if ( IN_DELETE_SELF & event->mask ) {
-		hit_delete_self[event->wd - 1]++;
-		num_delete_self++;
+		++w->hit_delete_self;
+		++num_delete_self;
 	}
 	if ( IN_UNMOUNT & event->mask ) {
-		hit_unmount[event->wd - 1]++;
-		num_unmount++;
+		++w->hit_unmount;
+		++num_unmount;
 	}
 	if ( IN_MOVE_SELF & event->mask ) {
-		hit_move_self[event->wd - 1]++;
-		num_move_self++;
+		++w->hit_move_self;
+		++num_move_self;
 	}
 
-	hit_total[event->wd - 1]++;
-	num_total++;
+	++w->hit_total;
+	++num_total;
 
 }
 
+int *stat_ptr(watch *w, int event)
+{
+	if ( IN_ACCESS == event )
+		return &w->hit_access;
+	if ( IN_MODIFY == event )
+		return &w->hit_modify;
+	if ( IN_ATTRIB == event )
+		return &w->hit_attrib;
+	if ( IN_CLOSE_WRITE == event )
+		return &w->hit_close_write;
+	if ( IN_CLOSE_NOWRITE == event )
+		return &w->hit_close_nowrite;
+	if ( IN_OPEN == event )
+		return &w->hit_open;
+	if ( IN_MOVED_FROM == event )
+		return &w->hit_moved_from;
+	if ( IN_MOVED_TO == event )
+		return &w->hit_moved_to;
+	if ( IN_CREATE == event )
+		return &w->hit_create;
+	if ( IN_DELETE == event )
+		return &w->hit_delete;
+	if ( IN_DELETE_SELF == event )
+		return &w->hit_delete_self;
+	if ( IN_UNMOUNT == event )
+		return &w->hit_unmount;
+	if ( IN_MOVE_SELF == event )
+		return &w->hit_move_self;
+	if ( 0 == event )
+		return &w->hit_total;
+	return 0;
+}
+
 /**
  * Get statistics by a particular watch descriptor.
  *
@@ -1241,42 +1497,13 @@ void record_stats( struct inotify_event const * event ) {
  *         enabled, or -1 if @a event or @a wd are invalid.
  */
 int inotifytools_get_stat_by_wd( int wd, int event ) {
-	niceassert( collect_stats, "stats requested but stats collection not "
-	                           "enabled" );
+	if (!collect_stats) return -1;
 
-	if ( wd < 0 ) return -1;
-
-	if ( IN_ACCESS == event )
-		return hit_access[wd - 1];
-	if ( IN_MODIFY == event )
-		return hit_modify[wd - 1];
-	if ( IN_ATTRIB == event )
-		return hit_attrib[wd - 1];
-	if ( IN_CLOSE_WRITE == event )
-		return hit_close_write[wd - 1];
-	if ( IN_CLOSE_NOWRITE == event )
-		return hit_close_nowrite[wd - 1];
-	if ( IN_OPEN == event )
-		return hit_open[wd - 1];
-	if ( IN_MOVED_FROM == event )
-		return hit_moved_from[wd - 1];;
-	if ( IN_MOVED_TO == event )
-		return hit_moved_to[wd - 1];
-	if ( IN_CREATE == event )
-		return hit_create[wd - 1];
-	if ( IN_DELETE == event )
-		return hit_delete[wd - 1];
-	if ( IN_DELETE_SELF == event )
-		return hit_delete_self[wd - 1];
-	if ( IN_UNMOUNT == event )
-		return hit_unmount[wd - 1];
-	if ( IN_MOVE_SELF == event )
-		return hit_move_self[wd - 1];
-
-	if ( 0 == event )
-		return hit_total[wd - 1];
-
-	return -1;
+	watch *w = watch_from_wd(wd);
+	if (!w) return -1;
+	int *i = stat_ptr(w, event);
+	if (!i) return -1;
+	return *i;
 }
 
 /**
@@ -1293,8 +1520,7 @@ int inotifytools_get_stat_by_wd( int wd, int event ) {
  *         is not a valid event.
  */
 int inotifytools_get_stat_total( int event ) {
-	niceassert( collect_stats, "stats requested but stats collection not "
-	                           "enabled" );
+	if (!collect_stats) return -1;
 	if ( IN_ACCESS == event )
 		return num_access;
 	if ( IN_MODIFY == event )
@@ -1343,6 +1569,9 @@ int inotifytools_get_stat_total( int event ) {
  *         the file specified by @a filename since stats collection was
  *         enabled, or -1 if the file is not being watched or @a event is
  *         invalid.
+ *
+ * @note The filename specified must always be the original name used to
+ *       establish the watch.
  */
 int inotifytools_get_stat_by_filename( char const * filename,
                                                 int event ) {
@@ -1371,6 +1600,7 @@ int isdir( char const * path ) {
 	static struct stat64 my_stat;
 
 	if ( -1 == lstat64( path, &my_stat ) ) {
+		if (errno == ENOENT) return 0;
 		fprintf(stderr, "Stat failed on %s: %s\n", path, strerror(errno));
 		return 0;
 	}
@@ -1386,7 +1616,9 @@ int isdir( char const * path ) {
  *         inotifytools_watch_files() and inotifytools_watch_recursively().
  */
 int inotifytools_get_num_watches() {
-	return num_watches;
+	int ret = 0;
+	rbwalk(tree_filename, get_num, (void*)&ret);
+	return ret;
 }
 
 /**
@@ -1612,7 +1844,8 @@ int inotifytools_snprintf( char * out, int size,
 	}
 
 	ind = 0;
-	for ( i = 0; i < strlen(fmt) && ind < size - 1; ++i ) {
+	for ( i = 0; i < strlen(fmt) &&
+	             (int)ind < size - 1; ++i ) {
 		if ( fmt[i] != '%' ) {
 			out[ind++] = fmt[i];
 			continue;
@@ -1634,7 +1867,7 @@ int inotifytools_snprintf( char * out, int size,
 
 		if ( ch1 == 'w' ) {
 			if ( filename ) {
-				strncpy( &out[ind], filename, MAX_STRLEN - ind );
+				strncpy( &out[ind], filename, size - ind );
 				ind += strlen(filename);
 			}
 			++i;
@@ -1643,7 +1876,7 @@ int inotifytools_snprintf( char * out, int size,
 
 		if ( ch1 == 'f' ) {
 			if ( eventname ) {
-				strncpy( &out[ind], eventname, MAX_STRLEN - ind );
+				strncpy( &out[ind], eventname, size - ind );
 				ind += strlen(eventname);
 			}
 			++i;
@@ -1652,7 +1885,7 @@ int inotifytools_snprintf( char * out, int size,
 
 		if ( ch1 == 'e' ) {
 			eventstr = inotifytools_event_to_str( event->mask );
-			strncpy( &out[ind], eventstr, MAX_STRLEN - ind );
+			strncpy( &out[ind], eventstr, size - ind );
 			ind += strlen(eventstr);
 			++i;
 			continue;
@@ -1675,7 +1908,7 @@ int inotifytools_snprintf( char * out, int size,
 				timestr[0] = 0;
 			}
 
-			strncpy( &out[ind], timestr, MAX_STRLEN - ind );
+			strncpy( &out[ind], timestr, size - ind );
 			ind += strlen(timestr);
 			++i;
 			continue;
@@ -1684,7 +1917,7 @@ int inotifytools_snprintf( char * out, int size,
 		// Check if next char in fmt is e
 		if ( i < strlen(fmt) - 2 && fmt[i+2] == 'e' ) {
 			eventstr = inotifytools_event_to_str_sep( event->mask, ch1 );
-			strncpy( &out[ind], eventstr, MAX_STRLEN - ind );
+			strncpy( &out[ind], eventstr, size - ind );
 			ind += strlen(eventstr);
 			i += 2;
 			continue;
@@ -1756,3 +1989,74 @@ int inotifytools_get_max_user_watches() {
 	if ( !read_num_from_file( WATCHES_SIZE_PATH, &ret ) ) return -1;
 	return ret;
 }
+
+/**
+ * Ignore inotify events matching a particular regular expression.
+ *
+ * @a pattern is a regular expression and @a flags is a bitwise combination of
+ * POSIX regular expression flags.
+ *
+ * On future calls to inotifytools_next_events() or inotifytools_next_event(),
+ * the regular expression is executed on the filename of files on which
+ * events occur.  If the regular expression matches, the matched event will be
+ * ignored.
+ */
+int inotifytools_ignore_events_by_regex( char const *pattern, int flags ) {
+	if (!pattern) {
+		if (regex) {
+			regfree(regex);
+			free(regex);
+			regex = 0;
+		}
+		return 1;
+	}
+
+	if (regex) { regfree(regex); }
+	else       { regex = (regex_t *)malloc(sizeof(regex_t)); }
+
+	int ret = regcomp(regex, pattern, flags | REG_NOSUB);
+	if (0 == ret) return 1;
+
+	regfree(regex);
+	free(regex);
+	regex = 0;
+	error = EINVAL;
+	return 0;
+}
+
+int event_compare(const void *p1, const void *p2, const void *config)
+{
+	if (!p1 || !p2) return p1 - p2;
+	char asc = 1;
+	int sort_event = (int)config;
+	if (sort_event == -1) {
+		sort_event = 0;
+		asc = 0;
+	} else if (sort_event < 0) {
+		sort_event = -sort_event;
+		asc = 0;
+	}
+	int *i1 = stat_ptr((watch*)p1, sort_event);
+	int *i2 = stat_ptr((watch*)p2, sort_event);
+	if (0 == *i1 - *i2) {
+		return ((watch*)p1)->wd - ((watch*)p2)->wd;
+	}
+	if (asc)
+		return *i1 - *i2;
+	else
+		return *i2 - *i1;
+}
+
+struct rbtree *inotifytools_wd_sorted_by_event(int sort_event)
+{
+	struct rbtree *ret = rbinit(event_compare, (void*)sort_event);
+	RBLIST *all = rbopenlist(tree_wd);
+	void const *p = rbreadlist(all);
+	while (p) {
+		void const *r = rbsearch(p, ret);
+		niceassert((int)(r == p), "Couldn't insert watch into new tree");
+		p = rbreadlist(all);
+	}
+	rbcloselist(all);
+	return ret;
+}
